'Mechanics of designations for interfaces and their implementations.'
from . import abstract
import abc
import collections
import weakref
import functools

class BoundIdentifier(tuple):
	"""
	An interface identifier bound to a particular Registry.
	"""
	__slots__ = ()

	@property
	def registry(self):
		return self[0]

	@property
	def identifier(self):
		return self[1]

	def hasinterface(self, cls):
		"""
		hasinterface(cls)

		Whether or not a given class provides the this referenced Interface.
		"""
		self.registry.hasinterface(cls, self.identifier)

	def isinstance(self, ob):
		"""
		isinstance(ob)

		Whether or not the given object's class has the specified interface.
		"""
		return self.hasinterface(ob.__class__)

	def isdefinition(self, cls):
		"""
		isdefinition(cls, iid)

		Whether or not a given class defines the given interface.
		"""
		return self.registry.isdefinition(cls, self.identifier)

	def issubface(self, *interfaces):
		return self.registry.issubface(self.identifier, *interfaces)

class Registry(object):
	"""
	Registry()

	Manages a set of interface and implementation associations.

	Normally, the instance at :py:obj:`visage.lib.registry` should be used,
	but there may be cases where an application desires to have its own
	universe.
	"""
	__slots__ = ('interfaces', 'implementations', 'data')

	def __init__(self,
		Set = weakref.WeakSet,
		VDictionary = weakref.WeakValueDictionary,
		KDictionary = weakref.WeakKeyDictionary
	):
		self.implementations = collections.defaultdict(Set)
		self.interfaces = VDictionary()
		self.data = KDictionary()

	def record_interface(self, cls, iid = ()):
		"""
		Register an interface as specifying the given interface identifier.
		"""
		register = cls.register # interfaces must be abc.ABCMeta instances
		iid = self.data[cls] = set(iid)
		for x in iid:
			# new interface registered,
			# register implementations with interface
			# xxx: throw warning on override
			self.interfaces[x] = cls
			for imp in self.implementations.get(x, ()):
				register(imp)
		return cls

	def record_implementation(self, cls, iid = ()):
		"""
		Register an implementation as implementing the given set of interface identifiers.
		"""
		iid = self.data[cls] = set(iid)
		for x in iid:
			# new implementation registered,
			# register with existing interfaces
			self.implementations[x].add(cls)
			ifc = self.interfaces.get(x)
			if ifc is not None:
				ifc.register(cls)
		return cls

	def interface(self, iid, partial = functools.partial):
		"""
		Decorator to add an interface to the registry.
		"""
		return partial(self.record_interface, iid = (iid,))

	def implementation(self, iid, partial = functools.partial):
		"""
		Decorator to add an implementation to the registry.
		"""
		return partial(self.record_implementation, iid = (iid,))

	def reference(*args, BoundIdentifier = BoundIdentifier):
		"""
		Returns a reference to an interface identifier bound to this registry.
		"""
		return BoundIdentifier(args)

	def hasinterface(self, cls, iid = None, type = type):
		"""
		hasinterface(cls, iid)

		Whether or not a given class provides the specified interface.
		"""
		##
		# This is the core test.
		#
		# In cases where an answer isn't immediately available,
		# it will go all the way to the end of the MRO to find
		# out whether or not a class has an interface.
		##
		data = self.data.get(cls)
		if data is not None and iid in data:
			# There's a registry entry for this particular class.
			return True

		# Even if this class has data, it might not have the full set
		# of IIDs, so we have to fall through and check ABC subclasses.

		# Try abc registry..
		if iid in self.interfaces:
			abc = self.interfaces[iid]
			if issubclass(cls, abc):
				# record the implementation class (caching sub. lookups)
				self.record_implementation(cls, iid)
				return True

		# Not in interfaces. Not in registry. Walk MRO
		mro = type.mro(cls)
		for cls in mro:
			data = self.data.get(cls)
			if data is not None and iid in data:
				# superclass noted implementation, cache this subclass.
				self.record_implementation(cls, iid)
				return iid in data

		return False

	def isinstance(self, ob, iid = None):
		"""
		isinstance(ob, iid)

		Whether or not the given object's class has the specified interface.
		"""
		if self.hasinterface(ob.__class__, iid):
			return True
		elif hasattr(ob, '__faces__'):
			return iid in ob.__faces__
		return False

	def isdefinition(self, cls, iid = None, ABC=abc.ABCMeta):
		"""
		isdefinition(cls, iid)

		Whether or not a given class defines the given interface.
		"""
		return isinstance(cls, ABC) and self.data.get(cls).issuperset(set((iid,)))

	def issubface(self, iid, *interfaces):
		"""
		issubface(iid, *interfaces)

		Whether the first IID is a sub-class of any of the given interfaces.

		Returns :py:obj:`None` when no interface class is present with the given `iid`.
		"""
		if iid in interfaces:
			return True

		if not iid in self.interfaces:
			return None
		Class = self.interfaces[iid]
		for y in interfaces:
			if self.hasinterface(Class, y):
				return True
		return False

	def analyze(self, Class, type = type):
		"""
		analyze(Class)

		Walk the MRO and return the set of *all* interfaces associated with the
		given Class.

		While strange, it is possible for results of the same function to change across
		calls.
		"""
		# Known IIDs
		get = self.data.get
		s = set()

		# Crawl the MRO of the class adding interfaces to the local set.
		for x in type.mro(Class):
			xiids = get(x)
			if xiids:
				s.update(xiids)
		return s

	def query(self, iid, subqueries, propertyclasses = (property, abc.abstractproperty)):
		"""
		query(iid, **subqueries)

		Query the implementation set for the given Interface Id with the properties
		identified by "subqueries".
		"""
		for x in self.implementations[iid]:
			for k,v in subqueries.items():
				# attribute on the implementation
				p = getattr(x, k)

				if isinstance(p, propertyclasses):
					if hasattr(p.fget, '__annotations__'):
						ref = p.fget.__annotations__.get('return')
					else:
						ref = None

					if ref is None:
						if v is not None:
							# qual could not be validated, fail out
							break
					elif isinstance(ref, BoundIdentifier):
						if not self.issubface(v, ref.identifier):
							break
				else:
					# attribute is a value, direct comparison.
					# we use annotated properties for subface checks.
					if v != p:
						break
			else:
				yield x

	def properties(self, instance, getattr = getattr, hasattr = hasattr):
		"""
		:returns: A list of tuples.
		:rtype: [(iid, attrname, return_annotation, prop)]

		Gather a list of properties according to the *interfaces* associated with
		the object.

		This method should be used in building out description graphs of a given instance.
		"""
		# gather all interface Id's
		ifes = self.analyze(instance.__class__)
		if not ifes:
			ifes = instance.__faces__

		# It's a component. Analyze the interfaces for properties.
		l = []

		for x in ifes:
			for name, attr in self.interfaces[x].__dict__.items():
				# add all abstractmethods that are properties
				if attr.__isabstractmethod__ and hasattr(attr, 'fget'):
					rtype = attr.fget.__annotations__.get('return')
					prop = getattr(instance, name)
					l.append((x, name, rtype, prop))

		return properties

def default():
	"""
	:rtype: :py:class:`Registry`

	Create a default registry.

	The following builtin ABCs are recorded with the lowercased identifiers:

	collections.Hashable
	 'hashable'
	collections.Iterator
	 'iterator'
	collections.Iterable
	 'iterable'
	collections.Sized
	 'sized'
	collections.Container
	 'container'
	collections.Callable
	 'callable'
	collections.Set
	 'set'
	collections.MutableSet
	 'mutableset'
	collections.Mapping
	 'mapping'
	collections.MutableMapping
	 'mutablemapping'
	collections.Sequence
	 'sequence'
	collections.MutableSequence
	 'mutablesequence'
	"""
	registry = Registry()

	# Record builtin ABCs.
	registry.record_interface(collections.Hashable, ('hashable',))
	registry.record_interface(collections.Iterator, ('iterator',))
	registry.record_interface(collections.Iterable, ('iterable',))
	registry.record_interface(collections.Sized, ('sized',))
	registry.record_interface(collections.Container, ('container',))
	registry.record_interface(collections.Callable, ('callable',))
	registry.record_interface(collections.Set, ('set',))
	registry.record_interface(collections.MutableSet, ('mutableset',))
	registry.record_interface(collections.Mapping, ('mapping',))
	registry.record_interface(collections.MutableMapping, ('mutablemapping',))
	registry.record_interface(collections.Sequence, ('sequence',))
	registry.record_interface(collections.MutableSequence, ('mutablesequence',))
	# Python doesn't have this in collections yet.
	registry.record_interface(abstract.Generator, ('generator',))

	return registry
